from bioTea.utils import tools


def test_tempdir_changer():
    import os

    prev_wd = os.getcwd()

    with tools.TempWorkdir("/tmp/banana"):
        new_wd = os.getcwd()

    restored_wd = os.getcwd()

    assert new_wd == "/tmp/banana"
    assert prev_wd == restored_wd


def test_ping():
    # I assume google is always up
    res = tools.ping("google.com")
    assert res is True


def test_making_geo_urls():
    codes = {
        "GPL": {
            "miniml": {"GPL"},
            "soft": "ftp://ftp.ncbi.nlm.nih.gov/geo/platforms/{obsid}/{id}/soft/",
            "suppl": "ftp://ftp.ncbi.nlm.nih.gov/geo/platforms/{obsid}/{id}/suppl/",
        },
        "GDS": {
            "miniml": "ftp://ftp.ncbi.nlm.nih.gov/geo/datasets/{obsid}/{id}/miniml/",
            "soft": "ftp://ftp.ncbi.nlm.nih.gov/geo/datasets/{obsid}/{id}/soft",
            "suppl": "ftp://ftp.ncbi.nlm.nih.gov/geo/datasets/{obsid}/{id}/suppl",
        },
        "GSM": {
            "miniml": "ftp://ftp.ncbi.nlm.nih.gov/geo/samples/{obsid}/{id}/miniml/",
            "soft": "ftp://ftp.ncbi.nlm.nih.gov/geo/samples/{obsid}/{id}/soft",
            "suppl": "ftp://ftp.ncbi.nlm.nih.gov/geo/samples/{obsid}/{id}/suppl",
        },
        "GSE": {
            "miniml": "ftp://ftp.ncbi.nlm.nih.gov/geo/series/{obsid}/{id}/miniml/",
            "soft": "ftp://ftp.ncbi.nlm.nih.gov/geo/series/{obsid}/{id}/soft",
            "suppl": "ftp://ftp.ncbi.nlm.nih.gov/geo/series/{obsid}/{id}/suppl",
        },
    }



import importlib.resources as pkg_resources
import logging
import shutil
from enum import Enum
from pathlib import Path
from typing import Optional

import typer
import yaml

import bioTea.pour as pour
from bioTea import resources
from bioTea.docker_wrapper import SpecialCommand, run_special_biotea_command
from bioTea.utils.strings import INTRO, TEA_LOGO
from bioTea.utils.tools import make_path_valid
from bioTea.wizard import interactive_metadata_to_biotea_box_options, wizard

log = logging.getLogger(__name__)

# CLI structure
# biotea
#   - info: Get version information and more.
#       - biotea: Get info about the biotea tool, its version (and DOI?)
#       - containers: Get available containers, as well as those locally installed.
#   - update: Check for container and tool updates.
#   - wizard: run the wizard
#   - retrieve: Retrieve (and format) data from GEO
#   - prepare
#       - affymetrix: Prep affymetrix data for analysis
#       - agilent: Prep agilent data for analysis
#   - analyze: Analyze with BioTEA box an expression file
#   - annotate: Annotate an expression matrix
#       - generate: Generate annotations for some organism.

cli_root = typer.Typer(no_args_is_help=True)
info = typer.Typer()
prepare = typer.Typer()
annotate = typer.Typer()

cli_root.add_typer(info, name="info")
cli_root.add_typer(
    prepare, name="prepare", help="Prepare raw expression data to an expression matrix"
)
cli_root.add_typer(annotate, name="annotations", help="Annotate expression data.")


@cli_root.callback()
def context_handler():
    log.debug(f"Starting bioTEA.")


@info.callback(invoke_without_command=True)
def generic_info(ctx: typer.Context):
    """Get information on the status of the tool."""
    print(TEA_LOGO)
    if ctx.invoked_subcommand:
        return

    # This is only printed if there is no subcommand invoked.
    print(INTRO)


@info.command(name="containers")
def info_containers_command():
    """Get information on the downloaded and available BioTEA box containers."""
    pour.info_containers()


@info.command(name="biotea")
def info_biotea_command():
    """Get information on the version of bioTEA."""
    # Add some free real estate
    print("\n")
    pour.info_biotea()


@info.command(name="versions")
def info_biotea_command(
    version: str = typer.Argument(
        ..., help="Version of the bioTEA box to get versions from."
    )
):
    """Get the R package versions used inside the bioTEA container"""
    run_special_biotea_command(SpecialCommand.versions, version=version)


@cli_root.command(name="update")
def update_tool_command(
    force: bool = typer.Option(
        False, help="Skip the confirmation prompt and update immediately"
    )
):
    """Check bioTEA and the container repos for updates.

    This command also updates the latest container, if needed.
    """
    pour.update_tool(yes=force)


@cli_root.command(name="wizard")
def run_wizard_command():
    """Run the bioTEA wizard.

    The wizard helps in setting up, running, and exploring a BioTEA box analysis.
    """
    wizard()


@cli_root.command(name="retrieve")
def retrieve_command(
    output_path: Path = typer.Argument(
        ..., help="Path to a folder that will contain the output"
    ),
    geo_id: str = typer.Argument(
        ..., help="GEO id that needs to be retrieved, e.g. GSE15471"
    ),
):
    """Retrieve data from GEO regarding a GEO series.

    Also helps setting the options for the BioTEA box analysis by providing a metadata file.
    """
    pour.retrieve(output_path=output_path, geo_id=geo_id)


@prepare.command(name="agilent")
def prepare_agilent_command(
    input_dir: Path = typer.Argument(
        ..., help="Path to the folder with the input files"
    ),
    output_file: Path = typer.Argument(..., help="Path to the output file"),
    grep_pattern: str = typer.Argument(
        "\.txt$", help="Pattern with which to find the files"
    ),
    version: str = typer.Option("latest", help="Specify BioTEA box container version"),
    log_name: Optional[str] = typer.Option(
        None, help="Specify BioTEA box log name for the run"
    ),
    remove_controls: bool = typer.Option(False, help="Remove control probes?"),
    plot_number: Optional[int] = typer.Option(
        None, help="Maximum number of plots to show"
    ),
    plot_size: str = typer.Option("12,5", help="Size of plots as 'width,height'"),
    use_png: bool = typer.Option(
        False, help="Generate .png files instead of pdf files (600 ppi resolution)."
    ),
    verbose: bool = typer.Option(False, help="Increase verbosity of BioTEA box logs"),
):
    """Prepare agilent expression data for analysis."""
    pour.prepare_agilent(
        input_dir=input_dir,
        output_file=output_file,
        grep_pattern=grep_pattern,
        version=version,
        log_name=log_name,
        remove_controls=remove_controls,
        plot_number=plot_number,
        plot_size=plot_size,
        use_png=use_png,
        verbose=verbose,
    )


@prepare.command(name="affymetrix")
def prepare_affymetrix_command(
    input_dir: Path = typer.Argument(
        ..., help="Path to the folder with the input files"
    ),
    output_file: Path = typer.Argument(..., help="Path to the output file"),
    version: str = typer.Option("latest", help="Specify BioTEA box container version"),
    log_name: Optional[str] = typer.Option(
        None, help="Specify BioTEA box log name for the run"
    ),
    remove_controls: bool = typer.Option(False, help="Remove control probes?"),
    plot_number: int = typer.Option(1e10, help="Maximum number of plots to show"),
    plot_size: str = typer.Option("12,5", help="Size of plots as 'width,height'"),
    use_png: bool = typer.Option(
        False, help="Generate .png files instead of pdf files (600 ppi resolution)."
    ),
    verbose: bool = typer.Option(False, help="Increase verbosity of BioTEA box logs"),
):
    """Prepare affymetrix expression data for analysis."""
    pour.prepare_affymetrix(
        input_dir=input_dir,
        output_file=output_file,
        version=version,
        log_name=log_name,
        remove_controls=remove_controls,
        plot_number=plot_number,
        plot_size=plot_size,
        use_png=use_png,
        verbose=verbose,
    )


@cli_root.command(name="analyze")
def run_biotea_box_analysis_command(
    options_path: Path = typer.Argument(..., help="Path to the options file"),
    output_dir: Path = typer.Argument(..., help="Path to the output directory"),
    input_file: Path = typer.Argument(..., help="Path to the input expression matrix"),
    version: str = typer.Option("latest", help="Specify BioTEA box container version"),
    log_name: Optional[str] = typer.Option(
        None, help="Specify BioTEA box log name for the run"
    ),
    verbose: bool = typer.Option(False, help="Increase verbosity of BioTEA box logs"),
):
    """Run Differential Gene Expression with BioTEA box."""
    pour.run_biotea_box_analysis(
        options_path=options_path,
        output_dir=output_dir,
        input_file=input_file,
        version=version,
        log_name=log_name,
        verbose=verbose,
    )

    log.info("BioTEA completed.")


@cli_root.command(name="initialize")
def init_with_options_command(
    path: Path = typer.Argument(
        ..., help="Path to the folder that has to be initialized"
    ),
    metadata: Optional[Path] = typer.Argument(
        None, help="Path to a metadata file to use to generate the config"
    ),
):
    """Generate an options file suitable for `biotea analize`

    If a metadata file is specified, it should be a metadata file like one that
    is produced by `biotea retrieve`, with a first column with sample names and
    a series of other columns with the sample variables.
    """
    make_path_valid(path, dir=True)
    if metadata:
        log.debug("Got a metadata path, so running interactively.")
        options = interactive_metadata_to_biotea_box_options(metadata)
        with (path / "bioTEA_run_options.yaml").open("w+") as outstream:
            yaml.dump(options, outstream, default_flow_style=False)
        return

    shutil.copy(
        pkg_resources.path(resources, "bioTEA_run_default_options.yaml"),
        path / "bioTEA_run_options.yaml",
    )
    log.info(f"Initialized options @ {path / 'bioTEA_run_options.yaml'}")


class ValidSpecies(str, Enum):
    human = "human"
    drosophila = "drosophila"
    mouse = "mouse"
    rat = "rat"
    bee = "apis"


@annotate.command(name="apply")
def annotate_file_command(
    target: Path = typer.Argument(
        ..., help="Path to the input expression matrix to annotate"
    ),
    output: Path = typer.Argument(..., help="Path to the annotated output path"),
    annotation_database: str = typer.Argument(
        "internal",
        help="Annotation database to use. Pass 'internal' to use the default human database. Otherwise, a path to the database file generated with `annotations generate`",
    ),
    version: str = typer.Option("latest", help="Specify BioTEA box container version"),
    log_name: Optional[str] = typer.Option(
        None, help="Specify BioTEA box log name for the run"
    ),
    verbose: bool = typer.Option(False, help="Increase verbosity of BioTEA box logs."),
):
    """Annotate some expression data or DEA output with annotation data."""
    pour.annotate_file(
        target=target,
        output=output,
        annotation_database=annotation_database,
        version=version,
        log_name=log_name,
        verbose=verbose,
    )


@annotate.command(name="generate", hidden=True)
def generate_annotations_command(
    target: Path = typer.Argument(
        ..., help="Path to the file where the annotations will be stored"
    ),
    organism: ValidSpecies = typer.Argument(
        ValidSpecies.human, help="Species to generate annotations for"
    ),
    verbose: bool = typer.Option(False, help="Increase verbosity of BioTEA box logs."),
):
    """Generate annotations to use with BioTEA box."""
    print("You got to a hidden command! This has not been implemented yet.")



import importlib.resources as pkg_resources
import itertools
import logging
import shutil
import string
from collections import Counter
from enum import Enum
from itertools import permutations
from math import floor
from pathlib import Path
from typing import Optional, Union

import pandas as pd
import typer
import yaml
from colorama import Fore
from fuzzywuzzy import fuzz
from tqdm import tqdm

from bioTea import pour, resources
from bioTea.utils.errors import BioTeaError, ErrorManager, RaiserHandler
from bioTea.utils.path_checker import is_pathname_valid
from bioTea.utils.strings import WIZARD_LOGO
from bioTea.utils.tools import (
    Replacer,
    ask_choices,
    contains_numbers,
    recursive_dict_update,
    user_input,
)

log = logging.getLogger(__name__)


class ValidManufacturers(Enum):
    def __str__(self) -> str:
        return str(self.value)

    agilent = "Agilent Technologies"
    affymetrix = "Affymetrix"


def glob_manufacturer(manufacturer_string) -> ValidManufacturers:
    tiers = {}
    for value in ValidManufacturers.__members__.values():
        tiers[str(value)] = fuzz.ratio(str(value), manufacturer_string)
    best_match_score = max(tiers.values())
    best_key = max(tiers, key=tiers.get)
    if best_match_score < 60:
        log.warn(
            f"Unsure about the manufacturer. I think it is {best_key}, but at a confidence of only {best_match_score}. Is '{manufacturer_string}' supported?"
        )

    return best_key


def sanitize_design_string(values: list) -> list[str]:
    # Character-only
    def contains_illegal(x):
        return any([y in ", -" for y in str(x)])

    if any(
        [contains_numbers(x) for x in values] + [contains_illegal(x) for x in values]
    ):
        log.warn(
            "There are illegal characters in the design string. Getting rid of them by replacing the factors."
        )
        replacer = Replacer(list(string.ascii_uppercase))
        sanitized_values = replacer.sanitize(values)
        log.info(
            "Replaced values: {}".format(
                ", ".join([f"{old} > {new}" for old, new in replacer.matches.items()])
            )
        )
        return sanitized_values
    log.debug("No need to sanitize input design string.")
    return values


def standardize_pairings(values: list) -> list[str]:
    # Number only
    log.info("Standardizing parings.")
    replacer = Replacer(itertools.count(0, 1))
    sanitized_values = replacer.sanitize(values)

    log.info(
        "New pairings: {}".format(
            ", ".join([f"{old} > {new}" for old, new in replacer.matches.items()])
        )
    )
    return sanitized_values


def interactive_metadata_to_biotea_box_options(
    metadata: Union[Path, pd.DataFrame],
    primary_var: Optional[str] = None,
    use_all_contrasts: bool = False,
) -> dict:
    log.debug("Generating a BioTEA_box_options.yaml file.")
    log.warn(
        "Due to limitations in parsing yaml data, the helpful comments in the "
        + "options will be lost by generating the file in this way. "
        + "Please refer to the manual directly for help "
        + "if you need to edit the file manually later."
    )

    if type(metadata) != pd.DataFrame:
        metadata = pd.read_csv(metadata)

    if "Sample_id" not in metadata.columns:
        typer.echo("Metadata file does not have a 'Sample_id' column. Aborting.")
        raise typer.Abort()

    metadata = metadata.sort_values(by=["Sample_id"])
    samples = ", ".join(list(metadata["Sample_id"]))
    log.debug(f"Using the following sample order: {samples}")

    meta_vars = [x for x in metadata.columns if x not in ["Sample_id"]]

    if any(metadata.applymap(lambda x: x in ["na", "null", None])):
        log.warning(
            "Detected some possible NAs or Null values in the metadata. This could cause errors later. Perhaps check the metadata first?"
        )

    if primary_var is not None:
        if primary_var in meta_vars:
            log.warn(
                "The specified primary variable is not valid. Falling back to prompting the user."
            )
            primary_var = None

    while primary_var is None:
        primary_var = ask_choices("Select the test variable", meta_vars)

        if not typer.confirm(f"Selection: {primary_var}. Is this correct?"):
            primary_var = None

    parsed_design_var = sanitize_design_string(metadata[primary_var])

    contrasts = None
    if len(set(parsed_design_var)) == 1:
        log.error(
            f"There is only one level in the variable {primary_var}. Cannot proceed with the analysis."
        )

    all_contrasts = [f"{x}-{y}" for x, y in permutations(set(parsed_design_var), 2)]

    if use_all_contrasts:
        log.info(f"Using all contrasts: {all_contrasts}")
        contrasts = all_contrasts

    if contrasts is not None and len(meta_vars) == 2:
        base = ask_choices("Select the 'control' variable", set(parsed_design_var))
        test = [x for x in meta_vars if x != base][0]
        contrasts = [f"{test}-{base}"]

    while contrasts is None:
        typer.echo(
            f"Specify the contrasts of interest, as 'test-control test-control ...'.\nE.g. '{all_contrasts[0]}'"
        )
        contrasts = ask_choices(
            "Please make a selection:", all_contrasts, accept_list=True
        )

    # -- Pairings
    all_other_vars = [x for x in meta_vars if x != primary_var]
    pairings = None
    if len(all_other_vars) > 0:
        typer.echo(
            f"Are there sample pairings (I.E. different samples but taken from the same patient)? If so, specify the variable that describes the pairings. If there are none, leave this empty:"
        )
        pairing_var = ask_choices("Please make a selection:", all_other_vars + [""])
        if pairing_var == "":
            pairings = None
        else:
            pairings = standardize_pairings(list(metadata[pairing_var]))
            counter = Counter(pairings)
            if len((eccessive_pairs := [x for x, y in counter.items() if y > 2])) > 0:
                log.warn(
                    f"There are one or more pairing(s) ({len(eccessive_pairs)}) that occur more than twice: {eccessive_pairs}.\n Are you sure this is the correct pairing variable? Maybe there are technical replicates in the study?"
                )
                typer.confirm("Continue anyway?", abort=True)

    # -- Batch effect
    all_other_vars = [x for x in meta_vars if x not in [primary_var, pairing_var]]
    batch_var = None
    if len(all_other_vars) > 0:
        typer.echo(
            f"Is there a variable that specifies the analysis batch? Specify it now. If not, leave this empty."
        )
        batch_var = ask_choices("Please make a selection:", all_other_vars + [""])
        if batch_var == "":
            batch_var = None

    all_other_vars = [
        x for x in meta_vars if x not in [primary_var, batch_var, pairing_var]
    ]
    other_vars = None
    if len(all_other_vars) > 0:
        typer.echo(
            f"Are there any other variables to account for? Specify them here, in a comma-separated list (e.g. var1, var2, var3...):"
        )
        other_vars = ask_choices(
            "Please make a selection:", all_other_vars + [""], accept_list=True
        )
        if other_vars == [""]:
            other_vars = None

    typer.echo("Recap:")
    typer.echo(f"\tPrimary variable to analize: {primary_var}")
    typer.echo(f"\tContrasts: {', '.join(contrasts)}")
    if batch_var:
        typer.echo(f"\tVariable to correct batches on: {batch_var}")
    else:
        typer.echo("\tNo batches to correct for.")
    if other_vars:
        typer.echo(f"\tOther variables to correct for: {', '.join(other_vars)}")
    else:
        typer.echo(f"\tNo other variables to take into account.")

    typer.confirm("\nIs this correct?", abort=True)

    log.info("Updating defaults with user data.")
    defaults = yaml.safe_load(
        pkg_resources.open_text(resources, "bioTEA_run_default_options.yaml")
    )
    log.debug("Loaded defaults.")

    if pairings:
        parsed_design_var = [f"{x}{y}" for x, y in zip(parsed_design_var, pairings)]
    new_expr_design = ", ".join(parsed_design_var)
    if batch_var:
        new_batches_design = ",".join(sanitize_design_string(metadata[batch_var]))
    else:
        new_batches_design = None
    if other_vars:
        new_extra_vars = {
            var: ", ".join(sanitize_design_string(metadata[var])) for var in other_vars
        }
    else:
        new_extra_vars = None

    defaults["design"]["experimental_design"] = new_expr_design
    log.debug(f"New design string: {new_expr_design}")
    defaults["design"]["contrasts"] = contrasts
    log.debug(f"New contrasts: {contrasts}")
    defaults["design"]["batches"] = new_batches_design
    log.debug(f"New batch string: {new_batches_design}")
    defaults["design"]["extra_limma_vars"] = new_extra_vars
    log.debug(f"New extra limma vars: {new_extra_vars}")

    return defaults


def interactive_general_options():
    show_snip = typer.confirm("Show data snippets during the analysis?", True)
    annotation_database = typer.confirm(
        "Use the internal (human) annotation database?", True
    )
    if not annotation_database:
        log.info(
            Fore.GREEN
            + "If you need to use a different annotation database, consider using `biotea annotations` instead, or edit the analysis file manually."
            + Fore.RESET
        )

    save_png = typer.confirm("Save png plots instead of pdf plots?", False)
    if save_png:
        png_resolution = typer.prompt("Png resolution (in pixels per inch)", 300)
    else:
        png_resolution = 300
    plot_width = typer.prompt("Plot width (in inches)", 16)
    plot_height = typer.prompt("Plot height (in inches)", 9)

    enumerate_plots = typer.confirm("Enumerate plots while saving them?", True)

    dry = typer.confirm(
        "Run in dryrun mode (this runs the analysis, but saves no output)?", False
    )
    renorm = typer.confirm(
        "Renormalize input data before analyzing (useful if the input data is really not homogeneous even after preparation)?",
        False,
    )
    convert_counts = typer.confirm(
        "Is the input data a count matrix (from RNA-seq)? Should the data be made continuous with `voom`?",
        False,
    )
    while True:
        limma = typer.confirm("Run the limma analysis?", True)
        rp = typer.confirm("Run the rankproduct analysis?", True)
        if limma or rp:
            break
        log.warning("Please run at least one between limma and rankprod.")

    return {
        "general": {
            "show_data_snippets": show_snip,
            "annotation_database": annotation_database,
            "plots": {
                "save_png": save_png,
                "plot_width": plot_width,
                "plot_height": plot_height,
                "png_resolution": png_resolution,
                "enumerate_plots": enumerate_plots,
            },
        },
        "switches": {
            "dryrun": dry,
            "renormalize": renorm,
            "convert_counts": convert_counts,
            "limma": limma,
            "rankproduct": rp,
        },
    }


def wizard_unhandled():
    typer.echo(WIZARD_LOGO)

    typer.echo(
        "Welcome to the bioTEA wizard. This helper will guide you through the whole bioTEA analysis."
    )
    typer.echo("You can exit at any time by pressing CTRL+C")
    typer.echo(
        Fore.YELLOW
        + "WARNING: The wizard is still experimental, and may produce errors. Use with caution."
    )
    typer.echo(
        "Select a folder (that will be created if non-existant) to use as a working directory:"
    )
    staging_path = user_input(
        "> ",
        is_pathname_valid,
        retry_prompt="The path seems invalid. Try again.",
    )
    log.debug(f"Got for path: {staging_path}")

    staging_path, temp_path = pour.stage_new_analysis(staging_path)

    typer.echo("Please specify the GEO id of the data. It should start with 'GSE':")
    geo_id = user_input(
        "> ",
        lambda x: x.startswith("GSE"),
        retry_prompt="That seems wrong. GEO series should start with 'GSE...'. Try again.",
    )

    series = pour.retrieve_geo_data(staging_path, geo_id, temp_dir=temp_path)

    new_options = interactive_metadata_to_biotea_box_options(series.generate_metadata())

    new_options = recursive_dict_update(new_options, interactive_general_options())
    log.debug(f"Final input dict: {new_options}")

    typer.echo(
        "We can start the analysis. First, we prepare the expression data to an expression frame."
    )

    best_platform = glob_manufacturer(series.platform.manufacturer)
    is_platform_correct = typer.confirm(
        f"The manufacturer label is {series.platform.manufacturer}. I think it is {best_platform}. Am I correct?",
        True,
    )
    if not is_platform_correct:
        log.error(
            "Cannot proceed. Prepare the data manually, then run `biotea initialize` followed by `biotea analyze yourself."
        )
        raise typer.Abort

    remove_controls = typer.confirm("Remove control probes from the input data?", True)
    if typer.confirm(
        "Reduce the number of diagnostic plots (to 10% of all samples)?", True
    ):
        plot_number = floor(len(series.samples) * 0.1)
    else:
        plot_number = len(series.samples)
    log.debug(f"Plot number set to be {plot_number}")

    for sample in tqdm(series.samples, "Unpacking samples..."):
        shutil.unpack_archive(
            sample.suppl_data_local_path, temp_path / "unpacked_raw_samples"
        )

    expression_file_path = staging_path / "expression_matrix.csv"
    if best_platform == ValidManufacturers.affymetrix:
        log.info("Preparing affymetrix data.")

        pour.prepare_affymetrix(
            temp_path / "unpacked_raw_samples",
            expression_file_path,
            remove_controls=remove_controls,
            plot_number=plot_number,
            plot_size=f"{new_options['general']['plots']['plot_width']},{new_options['general']['plots']['plot_height']}",
            use_png=new_options["general"]["plots"]["save_png"],
        )
    elif best_platform == ValidManufacturers.agilent:
        log.info("Preparing agilent data.")

        # Some magic to get the prefix
        suffix = Path(series.samples[0].suppl_data_local_path).suffix
        log.debug(f"Suffix automatically detected to be '{suffix}'")
        pour.prepare_agilent(
            temp_path / "unpacked_raw_samples",
            expression_file_path,
            # The \ is there to escape the .
            grep_pattern=f"\{suffix}",
            remove_controls=remove_controls,
            plot_number=plot_number,
            plot_size=f"{new_options['general']['plots']['plot_width']},{new_options['general']['plots']['plot_height']}",
            use_png=new_options["general"]["plots"]["save_png"],
        )

    log.info(
        f"Done preparing the input data. Prepared data is @ {staging_path / 'expression_matrix.csv'}"
    )

    log.info("Running container.")
    log.debug("Writing options file.")
    options_path = staging_path / "wizard_box_options.yaml"
    with options_path.open("w+") as fileout:
        yaml.dump(new_options, fileout)

    pour.run_biotea_box_analysis(
        options_path=options_path,
        output_dir=staging_path / "analysis",
        input_file=expression_file_path,
    )

    typer.echo(
        f"The analysis ended. Take a look inside {staging_path} for the output. Temporary files will be cleaned automatically."
    )
    log.info("Wizard completed.")


def wizard(*args, **kwargs):
    # Without this, the error manager tries to give args to the wizard,
    # which causes it to crash.
    def _wrapper(*args, **kwargs):
        wizard_unhandled()

    wizard_manager = ErrorManager(_wrapper)

    wizard_manager.add_handler(BioTeaError, RaiserHandler())

    wizard_manager.run(args, kwargs)



import logging
import os
import shutil
import tempfile
from pathlib import Path
from sys import exc_info
from typing import Optional, Tuple

import typer
from colorama import Fore

from bioTea import __version__
from bioTea.classes import GeoSample
from bioTea.docker_wrapper import (
    AnalyzeInterface,
    AnnotateInterface,
    PrepAffyInterface,
    PrepAgilInterface,
    get_all_versions,
    get_installed_versions,
    get_latest_version,
    pull_biotea_box_version,
    run_biotea_box,
)
from bioTea.utils.errors import ImageNotFoundError, SanityError
from bioTea.utils.strings import TEA_LOGO
from bioTea.utils.tools import (
    PathLike,
    download_ftp,
    make_geo_ftp,
    make_path_valid,
    parse_biotea_box_options,
)
from bioTea.utils.xml_parser import get_minimal_from_geo

log = logging.getLogger(__name__)


def stage_new_analysis(staging_path: PathLike) -> Tuple[Path, Path]:
    """Stage a new analysis folder, with an output and temp folder.

    Args:
        staging_path (PathLike): The path to the output folder

    Returns:
        Tuple[Path, Path]: A tuple with first element the (resolved) output path,
        and second element the path to the temporary folder.
    """
    log.info("Staging new analysis project...")
    log.debug(f"Resolving {staging_path}")
    staging_path = Path(staging_path).expanduser().resolve()
    if not staging_path.exists():
        os.makedirs(staging_path)
    log.debug(f"Staged {staging_path}.")
    log.info("Making a new temporary directory...")
    temp_dir = Path(tempfile.mkdtemp()).resolve()
    log.debug(f"Staged tempdir {temp_dir}")

    return (staging_path, temp_dir)


def retrieve_geo_data(
    output_folder: PathLike,
    geo_id: str,
    temp_dir: Optional[Path] = None,
    unpack_to_temp: bool = False,
):
    if temp_dir is None:
        staging_path, temp_dir = stage_new_analysis(output_folder)
    else:
        staging_path = make_path_valid(output_folder)

    geo_series = get_minimal_from_geo(geo_id, temp_dir)

    log.info("Retrieving sample raw data...")
    downloaded = download_ftp(
        make_geo_ftp(geo_id, "suppl"),
        staging_path / "raw_data",
        filter=lambda x: "_RAW" in x,
    )
    log.debug("Testing validity of downloaded archive...")
    # There should be just a single file
    if len(downloaded) > 1:
        log.warn("Downloaded more than one file. Using just the first one.")

    log.debug("Unpacking downloaded archive...")
    unpack_dir = temp_dir if unpack_to_temp else staging_path
    unpack_dir /= "unpacked_samples"
    log.debug(f"Unpacking dir is {unpack_dir}")
    shutil.unpack_archive(list(downloaded.values())[0], unpack_dir)
    unpacked_files = os.listdir(unpack_dir)

    log.debug(f"Unpacked {len(unpacked_files)} files: {unpacked_files}")

    log.debug("Sanity Check: Testing congruency with MINiML file...")
    miniml_samples = [
        sample.suppl_data_ftp.split("/")[-1] for sample in geo_series.samples
    ]
    if not all(miniml_sample in unpacked_files for miniml_sample in miniml_samples):
        log.error("Mismatching download and MINiML file. Aborting.")
        raise SanityError("Mismatching downloaded and MINiML files.")

    def add_realpath(sample: GeoSample, realpath: Path) -> GeoSample:
        sample.suppl_data_local_path = realpath
        return sample

    log.debug("Adding real paths to sample objects...")
    geo_series.samples = [
        add_realpath(sample, unpack_dir / sample.suppl_data_ftp.split("/")[-1])
        for sample in geo_series.samples
    ]

    log.info("Done retrieving project data.")
    return geo_series


def info_containers():
    """Get information on the downloaded and available BioTeaBox containers."""
    log.info("Getting container info...")
    local_versions = get_installed_versions()
    remote_versions = get_all_versions()

    c = lambda x: Fore.LIGHTGREEN_EX + str(x) + Fore.RESET
    col_remote_vers = [
        c(ver) if ver in local_versions else str(ver) for ver in remote_versions
    ]

    local_versions = [str(x) for x in local_versions]
    typer.echo(Fore.LIGHTBLUE_EX + "--- Container Info ---" + Fore.RESET)
    typer.echo("Locally installed: {}".format(", ".join(sorted(local_versions))))
    typer.echo("Remotely available: {}".format(", ".join(sorted(col_remote_vers))))
    typer.echo(
        Fore.LIGHTGREEN_EX
        + "Note: "
        + Fore.RESET
        + "Remote containers installed locally are highlighted in green."
    )
    typer.echo(Fore.LIGHTBLUE_EX + "----------------------" + Fore.RESET)


def info_biotea():
    """Get information on the version of bioTEA."""
    typer.echo(Fore.LIGHTBLUE_EX + "--- BioTEA Info ---" + Fore.RESET)
    typer.echo(Fore.LIGHTGREEN_EX + "Version: " + Fore.RESET + __version__)


def update_tool(yes: bool = False):
    """Check bioTEA and the container repos for updates.

    This command also updates the latest container, if needed.
    """
    # TODO: Once this is released on PyPA, implement checks for the tool update.
    log.info("Checking containers for updates...")
    try:
        if (latest := get_latest_version()) in get_installed_versions():
            log.info("Containers are up-to-date.")
            return
    except ImageNotFoundError:
        log.error(
            "No remote versions found, therefore there is no latest version. Cannot update."
        )
        return

    if not yes:
        do_update = typer.confirm(
            f"A new BioTeaBox version was found ({latest}). Update?",
        )
    else:
        log.debug("Skipped confirmation prompt")
        do_update = True
    if not do_update:
        log.debug("User abort.")
        return

    pull_biotea_box_version(latest)

    log.info("Done pulling new version.")


def retrieve(
    output_path: Path,
    geo_id: str,
):
    """Retrieve data from GEO regarding a GEO series.

    Also helps setting the options for the BioTeaBox analysis by providing a metadata file.
    """
    try:
        geo_series = retrieve_geo_data(output_folder=output_path, geo_id=geo_id)
    except Exception:
        log.error(
            "Failed to retrieve GEO data. Possibly, the MINiML file cannot be parsed correctly."
        )
        log.debug("Stack:", exc_info=exc_info())
        return

    log.info("Writing metadata...")
    with (output_path / "metadata.csv").open("w+") as fileout:
        geo_series.generate_metadata().to_csv(fileout, doublequote=True, index=False)

    log.info(f"Done retrieving data for {geo_id}.")


def prepare_agilent(
    input_dir: Path,
    output_file: Path,
    grep_pattern: str = "\.txt",
    version: str = "latest",
    log_name: Optional[str] = None,
    remove_controls: bool = False,
    plot_number: Optional[int] = None,
    plot_size: str = "12,5",
    use_png: bool = False,
    verbose: bool = False,
):
    """Prepare agilent expression data for analysis."""
    input_dir = make_path_valid(input_dir, dir=True)
    output_file = make_path_valid(output_file)

    if version == "latest":
        version = get_latest_version()

    try:
        plot_width, plot_height = [int(x) for x in plot_size.split(",")]
    except (ValueError, TypeError):
        log.error(f"Cannot parse {plot_size} as two numbers.")
        return

    if not version in ["bleeding"] and version not in get_all_versions():
        log.error(f"Invalid BioTeaBox version {version}")
        return

    if version not in get_installed_versions():
        pull_biotea_box_version(version)

    if plot_number is None:
        plot_number = 1_000_000

    args = {
        "output_file": output_file.name,
        "remove_controls": remove_controls,
        "n_plots": plot_number,
        "grep_pattern": grep_pattern,
        # Plot options
        "use_pdf": not use_png,
        "plot_width": plot_width,
        "plot_height": plot_height,
        "png_ppi": 600,
        "enumerate_plots": False,
    }

    response = run_biotea_box(
        "prepagil",
        arguments=args,
        interface=PrepAgilInterface,
        input_anchor=input_dir,
        output_anchor=output_file.parent,
        log_anchor=output_file.parent,
        version=version,
        console_level="debug" if verbose else "info",
        logfile_level="debug",
        log_name=log_name or "auto",
    )

    return response


def prepare_affymetrix(
    input_dir: Path,
    output_file: Path,
    version: str = "latest",
    log_name: Optional[str] = None,
    remove_controls: bool = False,
    plot_number: int = None,
    plot_size: str = "12,5",
    use_png: bool = False,
    verbose: bool = False,
):
    """Prepare affymetrix expression data for analysis."""
    input_dir = make_path_valid(input_dir, dir=True)
    output_file = make_path_valid(output_file)

    if version == "latest":
        version = get_latest_version()

    try:
        plot_width, plot_height = [int(x) for x in plot_size.split(",")]
    except (ValueError, TypeError):
        log.error(f"Cannot parse {plot_size} as two numbers.")
        return

    if not version in ["bleeding"] and version not in get_all_versions():
        log.error(f"Invalid BioTeaBox version {version}")
        return

    if version not in get_installed_versions():
        pull_biotea_box_version(version)

    if plot_number is None:
        plot_number = 1_000_000

    args = {
        "output.file": output_file.name,
        "remove.controls": remove_controls,
        "n_plots": plot_number,
        # Plot options
        "use_pdf": not use_png,
        "plot_width": plot_width,
        "plot_height": plot_height,
        "png_ppi": 600,
        "enumerate_plots": False,
    }

    response = run_biotea_box(
        "prepaffy",
        arguments=args,
        interface=PrepAffyInterface,
        input_anchor=input_dir,
        output_anchor=output_file.parent,
        log_anchor=output_file.parent,
        version=version,
        console_level="debug" if verbose else "info",
        logfile_level="debug",
        log_name=log_name or "auto",
    )

    return response


def run_biotea_box_analysis(
    options_path: Path,
    output_dir: Path,
    input_file: Path,
    version: str = "latest",
    log_name: Optional[str] = None,
    verbose: bool = False,
):
    """Run Differential Gene Expression with BioTeaBox."""
    print(TEA_LOGO)
    log.info(f"Biotea version {__version__}")

    options_path = make_path_valid(options_path)
    output_dir = make_path_valid(output_dir)
    input_file = make_path_valid(input_file)

    if version == "latest":
        version = get_latest_version()

    if not version in ["bleeding"] and version not in get_all_versions():
        log.error(f"Invalid BioTeaBox version {version}")
        return

    if version not in get_installed_versions():
        pull_biotea_box_version(version)

    log.info("Parsing options...")
    args = parse_biotea_box_options(options_path)

    log.debug("Adding input.file argument...")
    args["input.file"] = input_file.name

    response = run_biotea_box(
        "analyze",
        arguments=args,
        interface=AnalyzeInterface,
        input_anchor=input_file.parent,
        output_anchor=output_dir,
        log_anchor=output_dir,
        version=version,
        console_level="debug" if verbose else "info",
        logfile_level="debug",
        log_name=log_name or "auto",
    )

    return response


def annotate_file(
    target: Path,
    output: Path,
    annotation_database: str = "internal",
    version: str = "latest",
    log_name: Optional[str] = None,
    verbose: bool = False,
):
    """Annotate some expression data or DEA output with annotation data."""
    output = make_path_valid(output)

    if annotation_database != "internal":
        # TODO: Remove this once `annotations generate` gets implemented.
        log.warn(
            "Setting the annotation database to anything other than `internal` is currently not supported. Setting it to `internal` again. See issue *** on GitHub."
        )
        annotation_database = "internal"

    if annotation_database == "internal":
        annotation_database = True

    run_biotea_box(
        command="annotation",
        arguments={
            "expression_data_path": target.name,
            "output_path": output.name,
            "database_name": annotation_database,
        },
        interface=AnnotateInterface,
        input_anchor=target.parent,
        output_anchor=output.parent,
        log_anchor=output.parent,
        version=version,
        console_level="debug" if verbose else "info",
        logfile_level="debug",
        log_name=log_name or "auto",
    )



import re
from itertools import zip_longest

from colorama import Fore

# 7-bit C1 ANSI sequences
ansi_escape = re.compile(
    r"""
    \x1B  # ESC
    (?:   # 7-bit C1 Fe (except CSI)
        [@-Z\\-_]
    |     # or [ for CSI, followed by a control sequence
        \[
        [0-?]*  # Parameter bytes
        [ -/]*  # Intermediate bytes
        [@-~]   # Final byte
    )
""",
    re.VERBOSE,
)


def strip_colors(string):
    return ansi_escape.sub("", string)


def combine_linewise(a, b, padding="", strip=False, align_bottom=False, fix_len=True):
    lines_a = a.split("\n")
    lines_b = b.split("\n")

    if align_bottom:
        lines_a = list(reversed(lines_a))
        lines_b = list(reversed(lines_b))

    def max_len(lines):
        return max([len(strip_colors(x)) for x in lines])

    if max_len(lines_a) > max_len(lines_b) and fix_len:
        fill = " " * max_len(lines_a)
    elif max_len(lines_b) > max_len(lines_a) and fix_len:
        fill = " " * max_len(lines_b)
    else:
        fill = ""

    result = []
    for line_a, line_b in zip_longest(lines_a, lines_b, fillvalue=fill):
        if strip:
            line_a, line_b = line_a.strip(), line_b.strip()
        line_a = line_a + padding
        result.append(line_a + line_b)

    if align_bottom:
        result = reversed(result)

    return "\n".join(result)


def make_square(logo, side="left"):
    assert side in ["left", "right"]
    lines = logo.split("\n")
    longest = max([len(strip_colors(x)) for x in lines])

    res = []
    for line in lines:
        padding = " " * (longest - len(strip_colors(line)))

        if side == "left":
            res.append(line + padding)
        elif side == "right":
            res.append(padding + line)

    return "\n".join(res)


def break_long_lines(text: str, max_len: int = 80) -> str:
    """This breaks very long lines in `text` to `max_len`.

    Only breaks at spaces.
    """
    lines = text.split("\n")
    new_lines = []

    for line in lines:
        # If the line is already ok, leave it as-is
        if len(line) <= max_len:
            new_lines.append(line)
            continue

        # If not, we have work to do
        words = line.split(" ")
        new_line = ""
        for word in words:
            if len(f"{new_line} {word}") > (max_len + 1):
                # The lstrip is there to remove the extra space at the start
                # the same goes for the +1 above
                new_lines.append(new_line.lstrip())
                new_line = ""
            new_line += f" {word}"
        new_lines.append(new_line.lstrip())

    new_text = "\n".join(new_lines)

    return new_text


TEA = """                                                    __/\__
             ;,'                               . _  \\\\''//
     _o_    ;:;' __    _     _______________   -( )-/_||_\\
 ,-.'---`.__ ;  / /_  (_)___/_  __/ ____/   |   .'. \_()_/
((j`=====',-'  / __ \/ / __ \/ / / __/ / /| |    |   | . \\
 `-\     /    / /_/ / / /_/ / / / /___/ ___ |    ϕ---| .  \\
    `-=-'    /_.___/_/\____/_/ /_____/_/  |_|   .'. ,\_____'.
                              W I Z A R D
"""

TEAPOT = (Fore.RESET + "\n").join(
    [
        Fore.LIGHTBLACK_EX + "             ;,'",
        Fore.LIGHTRED_EX + "     _o_    " + Fore.LIGHTBLACK_EX + ";:;' ",
        Fore.LIGHTRED_EX + " ,-.'---`.__ " + Fore.LIGHTBLACK_EX + ";  ",
        Fore.LIGHTRED_EX + "((j`=====',-'  ",
        Fore.LIGHTRED_EX + " `-\     /    ",
        Fore.LIGHTRED_EX + "    `-=-'    ",
    ]
) + Fore.RESET

BIOTEA = (Fore.RESET + "\n").join(
    [
        "    __    _     " + Fore.LIGHTGREEN_EX + "_______________ ",
        "   / /_  (_)___" + Fore.LIGHTGREEN_EX + "/_  __/ ____/   |",
        "  / __ \/ / __ \\" + Fore.LIGHTGREEN_EX + "/ / / __/ / /| |",
        " / /_/ / / /_/ " + Fore.LIGHTGREEN_EX + "/ / / /___/ ___ |",
        "/_.___/_/\____" + Fore.LIGHTGREEN_EX + "/_/ /_____/_/  |_|",
    ]
)

BIOTEA_S = (Fore.RESET + "\n").join(
    [
        "__    _     " + Fore.LIGHTGREEN_EX + "_______________ ",
        "/ /_  (_)___" + Fore.LIGHTGREEN_EX + "/_  __/ ____/   |",
        "/ __ \/ / __ \\" + Fore.LIGHTGREEN_EX + "/ / / __/ / /| |",
        "/ /_/ / / /_/ " + Fore.LIGHTGREEN_EX + "/ / / /___/ ___ |",
        "/_.___/_/\____" + Fore.LIGHTGREEN_EX + "/_/ /_____/_/  |_|",
    ]
) + Fore.RESET

WIZARD = (Fore.RESET + "\n").join(
    [
        Fore.LIGHTYELLOW_EX + "     __/\__",
        Fore.LIGHTCYAN_EX + ". _ " + Fore.LIGHTYELLOW_EX + " \\\\''//",
        Fore.LIGHTCYAN_EX + "-( )-" + Fore.LIGHTBLACK_EX + "/_||_\\",
        Fore.LIGHTCYAN_EX + " .'. " + Fore.LIGHTBLACK_EX + "\_()_/",
        "  |   " + Fore.LIGHTRED_EX + "| . \\",
        Fore.LIGHTBLACK_EX + "  ϕ" + Fore.LIGHTRED_EX + "---| .  \\",
        " .'. " + Fore.LIGHTRED_EX + ",\_____'.",
    ]
) + Fore.RESET

WIZARD_WORD = Fore.LIGHTCYAN_EX + "                          W I Z A R D" + Fore.RESET

TEA_LOGO = combine_linewise(
    TEAPOT, BIOTEA_S, strip=False, align_bottom=True, fix_len=False
)
WIZARD_LOGO = (
    combine_linewise(
        make_square(TEA_LOGO), WIZARD, strip=False, align_bottom=True, padding=" "
    )
    + "\n"
    + WIZARD_WORD
)


INTRO = break_long_lines(
    """
BioTEA, where Tea is short for Transcript Enrichment Analysis, is a pipeline for Differential Gene expression Analysis with microarray and RNA-seq data.

It can download, preprocess and perform DEAs quickly, easily and in a reproducible way from the command line.

All commands (including the bare "biotea") have an "--help" option to get a description and more information on the possible options of the command. You are encouraged to use it to learn more for each command.

If you use BioTEA for your research, please cite our publication. Learn more at the README on GitHub.

For more information, try "biotea info biotea" and "biotea info containers".
"""
)



import logging
import traceback
from abc import ABC, abstractmethod
from typing import Callable

log = logging.getLogger(__name__)


# < Custom BioTea errors > #
class BioTeaError(Exception):
    """Generic class from which all BioTea Errors inherit from."""


class InvalidGeoId(BioTeaError):
    """Error raised when an input GEO id is invalid or not found."""

    pass


class UnsupportedChip(BioTeaError):
    """Raised when the chip ID is unsupported."""

    pass


class SanityError(BioTeaError):
    """Raised when a sanity check fails."""

    pass


class ImageNotFoundError(BioTeaError):
    """Raised when remote or local images are not found."""

    pass


class InvalidPathError(BioTeaError):
    """Raised when a given path is invalid."""

    pass


class ContainerExitError(BioTeaError):
    """Raised when the biotea-box exits with an error"""

    pass


# --- #


class ErrorHandler(ABC):
    """Abstract class from which error handlers can be specified."""

    error: Exception = None

    # I am not 100% sure this is the correct way to use this, but it fails
    # if "error" is not defined, so I'm happy.
    def __init_subclass__(cls, /, error, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.error = error

    @abstractmethod
    def handle(error: Exception, invoker: Callable, trace: str, inputs: dict) -> None:
        pass


class ErrorManager:
    """Class to run a fuction and handle errors that it raises.

    Since the calling function is passed to the handle, allows re-running the
    fuction again after the error is handled.
    """

    executable = None

    def __init__(self, executable: Callable) -> None:
        self.handlers = dict()
        self.executable = executable

    def add_handler(self, exception: Exception, handler: ErrorHandler) -> None:
        if exception in self.handlers.keys():
            log.warn(f"Overwriting handler for {exception}.")

        self.handlers.update({exception: handler})

    def run(self, *args, **kwargs):
        try:
            self.executable(*args, **kwargs)
        except Exception as e:
            if e in self.handlers.keys():
                target = e
            elif any(
                (
                    compatible_handlers := [
                        e for x in self.handlers.keys() if issubclass(x, type(e))
                    ]
                )
            ):
                if sum(compatible_handlers) > 1:
                    log.warning(
                        f"Multiple equally satisfactory handlers found ({compatible_handlers}). Using the last one added."
                    )
                target = compatible_handlers[-1]
            else:
                log.warn("No compatible handlers found. Raising the error.")
                raise e

            self.handlers[target].handle(
                error=e,
                invoker=self.executable,
                trace=traceback.format_exc(),
                inputs={"args": args, "kwargs": kwargs},
            )


class RaiserHandler(ErrorHandler, error=BioTeaError):
    """Handle that just raises the error. The 'non-handle'."""

    def handle(error: Exception, invoker: Callable, trace: str, inputs: dict) -> None:
        raise error


class UnsupportedOptionError(Exception):
    pass



"""bioTEA - Transcript Expression Analysis

A pipeline for processing transcription data to get differential gene expression.
"""
import collections
import importlib.resources as pkg_resources
import logging
import os
from copy import copy
from functools import reduce
from logging import StreamHandler
from logging.handlers import RotatingFileHandler
from pathlib import Path

import yaml
from colorama import Back, Fore, Style, init

from bioTea import resources
from bioTea.utils.errors import UnsupportedOptionError

init(autoreset=True)

__version__ = "1.0.2"
# Set what is exported by the __init__
__all__ = ["__version__", "OPTIONS"]

# Parse local options for the tool.
DEFAULT_OPTIONS = yaml.safe_load(
    pkg_resources.open_text(resources, "default_options.yml")
)


# Stolen from stackoverflow and modded
def recursive_options_update(d, u):
    for k, v in u.items():
        if k not in d:
            raise UnsupportedOptionError(f"Custom option {k} is not supported.")
        if isinstance(v, collections.abc.Mapping):
            d[k] = recursive_options_update(d.get(k, {}), v)
        else:
            d[k] = v
    return d


_possible_option_paths = [
    Path("~/.bioTEA/config.yaml"),
    Path("~/.bioTEA/config.yml"),
    Path("~/.config/bioTEA/config.yaml"),
    Path("~/.config/bioTEA/config.yml"),
]
_all_local_opts = []
for path in _possible_option_paths:
    if not path.exists():
        continue
    with path.open("r") as file:
        _all_local_opts.append(yaml.safe_load(file))
OPTIONS = reduce(recursive_options_update, _all_local_opts, copy(DEFAULT_OPTIONS))


class ColorFormatter(logging.Formatter):
    # Change this dictionary to suit your coloring needs!
    COLORS = {
        "WARNING": Fore.YELLOW,
        "ERROR": Fore.RED,
        "DEBUG": Style.BRIGHT + Fore.MAGENTA,
        "INFO": Fore.GREEN,
        "CRITICAL": Style.BRIGHT + Fore.RED,
    }

    def format(self, record):
        reset = Fore.RESET + Back.RESET + Style.NORMAL
        color = self.COLORS.get(record.levelname, "")
        if color:
            record.name = Style.BRIGHT + Fore.BLACK + record.name + reset
            if record.levelname != "INFO":
                record.msg = color + record.msg + reset
            record.levelname = color + record.levelname + reset
        return logging.Formatter.format(self, record)


# Setup logging
log = logging.getLogger("bioTea")  # Keep this at the module level name
log.setLevel(logging.DEBUG)
log.propagate = False
# Keep this at DEBUG - set levels in handlers themselves

format = "%(asctime)s [%(levelname)s] %(name)s: %(message)s"
file_formatter = logging.Formatter(format)
console_formatter = ColorFormatter(format)

_LOG_PATH = (Path(OPTIONS["log_folder"]) / "bioTEA.log").expanduser().resolve()

if not _LOG_PATH.parent.exists():
    os.makedirs(_LOG_PATH.parent)

file_h = RotatingFileHandler(
    filename=Path(_LOG_PATH),
    encoding="utf-8",
    mode="a+",
    maxBytes=1e5,
    backupCount=5,
)
file_h.setFormatter(file_formatter)
file_h.setLevel(logging.DEBUG)
stream_h = StreamHandler()
stream_h.setFormatter(console_formatter)
stream_h.setLevel(logging.DEBUG)

log.addHandler(file_h)
log.addHandler(stream_h)



import os
import subprocess
from pathlib import Path

import pytest

FIXTURE_DIR = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    "test_files",
)


def test_biotea_installed():
    subprocess.run(["biotea", "--help"]).check_returncode()


class TestBioTEACommands:
    def test_biotea_info(self):
        # Test that all commands work and can be accessed.
        subprocess.run(["biotea", "info"]).check_returncode()


def contains_logs(dir):
    return "biotea_log.log" in [path.name for path in Path(dir).iterdir()]


def contains_output(dir):
    # Are there files in this folder?
    # It's a pretty weak assertion, yeah, but I don't have time for anything better.
    return len(list(Path(dir).iterdir())) > 2


# Some terrible integration tests, just so we have *some* testing capabilities.
@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "Fake_options_files/fake_expression_run.yaml"),
    os.path.join(FIXTURE_DIR, "fake_expression_matrix.csv"),
)
def test_generic_run(tmpdir, datafiles):
    res = subprocess.run(
        [
            "biotea",
            "analyze",
            "--log-name",
            "biotea_log",
            "--version",
            "bleeding",
            os.path.join(datafiles, "fake_expression_run.yaml"),
            tmpdir,
            os.path.join(datafiles, "fake_expression_matrix.csv"),
        ],
        capture_output=True,
    )
    assert res.returncode == 0, f"Command failed: {res.stderr}"
    # Test the presence of the logs
    assert contains_logs(tmpdir)


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "Fake_options_files/fake_expression_run_wet.yaml"),
    os.path.join(FIXTURE_DIR, "fake_expression_matrix.csv"),
)
def test_generic_run_with_plots(tmpdir, datafiles):
    res = subprocess.run(
        [
            "biotea",
            "analyze",
            "--version",
            "bleeding",
            "--log-name",
            "biotea_log",
            os.path.join(datafiles, "fake_expression_run_wet.yaml"),
            tmpdir,
            os.path.join(datafiles, "fake_expression_matrix.csv"),
        ],
        capture_output=True,
    )
    assert res.returncode == 0, f"Command failed: {res.stderr}"
    # Test the presence of the logs and plots and outputs
    assert contains_logs(tmpdir)
    assert contains_output(tmpdir)
    assert (Path(tmpdir) / "analyze Figures").exists()
    # Did (at least one) plot get saved?
    assert len(list((Path(tmpdir) / "analyze Figures").iterdir())) > 0
    # Test that the correspondence table got saved
    assert (Path(tmpdir) / "correspondence_table.csv").exists()


@pytest.mark.datafiles(
    os.path.join(
        FIXTURE_DIR, "Fake_options_files/fake_expression_run_with_annots.yaml"
    ),
    os.path.join(FIXTURE_DIR, "fake_expression_matrix_annotatable.csv"),
)
def test_generic_run_with_annotations(tmpdir, datafiles):
    res = subprocess.run(
        [
            "biotea",
            "analyze",
            "--version",
            "bleeding",
            "--log-name",
            "biotea_log",
            os.path.join(datafiles, "fake_expression_run_with_annots.yaml"),
            tmpdir,
            os.path.join(datafiles, "fake_expression_matrix_annotatable.csv"),
        ],
        capture_output=True,
    )
    assert res.returncode == 0, f"Command failed: {res.stderr}"
    # Test the output files actually have the annotation columns
    assert contains_logs(tmpdir)
    assert contains_output(tmpdir)
    assert (Path(tmpdir) / "analyze Figures").exists()
    with (Path(tmpdir) / "Limma - DEG Table treated-control.csv").open("r") as testfile:
        header = testfile.readline()
    # The double quotes is due to csv encoding.
    assert '"SYMBOL"' in header.split(",")


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "Fake_options_files/fake_counts_run.yaml"),
    os.path.join(FIXTURE_DIR, "fake_count_matrix.csv"),
)
def test_generic_counts_run(tmpdir, datafiles):
    res = subprocess.run(
        [
            "biotea",
            "analyze",
            "--version",
            "bleeding",
            "--log-name",
            "biotea_log",
            os.path.join(datafiles, "fake_counts_run.yaml"),
            tmpdir,
            os.path.join(datafiles, "fake_count_matrix.csv"),
        ],
        capture_output=True,
    )
    assert res.returncode == 0, f"Command failed: {res.stderr}"
    # Test the presence of the logs
    assert contains_logs(tmpdir)


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DIR, "Fake_options_files/fake_counts_run.yaml"),
    os.path.join(FIXTURE_DIR, "fake_count_matrix_unsorted.csv"),
)
def test_generic_unsorted_cols(tmpdir, datafiles):
    res = subprocess.run(
        [
            "biotea",
            "analyze",
            "--version",
            "bleeding",
            os.path.join(datafiles, "fake_counts_run.yaml"),
            tmpdir,
            os.path.join(datafiles, "fake_count_matrix_unsorted.csv"),
        ],
        capture_output=True,
    )
    assert res.returncode == 0, f"Command failed: {res.stderr}"



"""xml parsing to get useful objects"""
# To annotate the class with itself
from __future__ import annotations

import logging
import os
import shutil
from pathlib import Path
from typing import BinaryIO

import xmltodict

from bioTea.classes import GeoPlatform, GeoSample, GeoSeries
from bioTea.utils.tools import PathLike, contains_all, download_ftp, make_geo_ftp

log = logging.getLogger(__name__)


class XmlMinimlExtractor:
    def __init__(self, xml_bytes: BinaryIO) -> None:
        self.xdict = xmltodict.parse(xml_bytes)["MINiML"]

    def extract_authors(self) -> list[str]:
        raw_authors = self.xdict["Contributor"]
        authors = []
        for author in raw_authors:
            if "Person" in author.keys():
                authors.append(
                    "{} {}".format(author["Person"]["First"], author["Person"]["Last"])
                )

        return authors

    def extract_samples(self) -> list[GeoSample]:
        raw_samples = self.xdict["Sample"]
        sample_objects = []
        for sample in raw_samples:
            sample_obj = GeoSample.sample_from_xmldict(sample)
            sample_objects.append(sample_obj)

        log.debug("Sanity Check: Do all samples have the same conditions?")
        for i, sample in enumerate(sample_objects):
            if not contains_all(
                list(sample_obj.conditions.keys()), list(sample.conditions.keys())
            ):
                extra_keys = set(sample.conditions.keys()) - set(
                    sample_obj.conditions.keys()
                )
                log.warn(
                    "Failed sanity check: Not all samples have the same conditions. "
                    f"(Extra keys: {extra_keys}) "
                    "Removing the extra keys."
                )
                for key in extra_keys:
                    log.debug(f"Popping {key}..")
                    sample.conditions.pop(key)

                log.debug(f"Dict after popping: {sample.conditions.keys()}")
                sample_objects[i] = sample

        return sample_objects

    def extract_platform(self) -> GeoPlatform:
        return GeoPlatform(
            accession=self.xdict["Platform"]["@iid"],
            manufacturer=self.xdict["Platform"]["Manufacturer"],
            title=self.xdict["Platform"]["Title"],
        )

    def extract_id(self) -> str:
        return self.xdict["Series"]["@iid"]


# Have xmltodict do the heavy lifting
def miniml_xml_to_series(xml_path: PathLike) -> GeoSeries:
    # Follow the paths to make the objects
    with Path(xml_path).open("rb") as bytes:
        extractor = XmlMinimlExtractor(bytes)

    project = GeoSeries(
        accession=extractor.extract_id(),
        authors=extractor.extract_authors(),
        platform=extractor.extract_platform(),
        samples=extractor.extract_samples(),
    )

    return project


def get_minimal_from_geo(
    geo_id: str, download_dir: PathLike, cleanup: bool = False
) -> GeoSeries:
    """Generate a GeoSeries object for a certain GEO Id.

    Uses the miniml file to generate the object.

    Args:
        geo_id (str): The ID of the wanted series.
        download_dir (PathLike): (Temporary) directory to download to.
        cleanup (bool, optional): If True, downloaded files are deleted once the fuction ends. Defaults to False.

    Returns:
        GeoSeries: The GeoSeries object for the specified series.
    """
    download_dir = Path(download_dir)
    log.info(f"Retrieving GEO MINiML file for {geo_id}")

    ftp_url = make_geo_ftp(geo_id, "miniml")
    downloaded = download_ftp(ftp_url, download_dir)
    # There should be just a single file, the zipped file with the data
    output_folder = download_dir / "minimal_files"
    os.makedirs(output_folder)
    # We downloaded only one file (hopefully)
    # TODO: this seems a hack
    shutil.unpack_archive(list(downloaded.values())[0], output_folder)

    unzipped_files = os.listdir(output_folder)
    file = [Path(x) for x in unzipped_files if x.endswith("_family.xml")][0]

    series = miniml_xml_to_series(output_folder / file)

    if cleanup:
        shutil.rmtree(output_folder)
        os.remove(downloaded)

    return series



# Path checking is a nightmare.
# I stole here the answer for this issue from stackoverflow:
# https://stackoverflow.com/questions/9532499/check-whether-a-path-is-valid-in-python-without-creating-a-file-at-the-paths-ta
# And I hope for the best.

# Please don't judge me

import errno
import os
import sys
import tempfile

# Sadly, Python fails to provide the following magic number for us.
ERROR_INVALID_NAME = 123
"""
Windows-specific error code indicating an invalid pathname.

See Also
----------
https://docs.microsoft.com/en-us/windows/win32/debug/system-error-codes--0-499-
    Official listing of all such codes.
"""


def is_pathname_valid(pathname: str) -> bool:
    """
    `True` if the passed pathname is a valid pathname for the current OS;
    `False` otherwise.
    """
    # If this pathname is either not a string or is but is empty, this pathname
    # is invalid.
    try:
        if not isinstance(pathname, str) or not pathname:
            return False

        # Strip this pathname's Windows-specific drive specifier (e.g., `C:\`)
        # if any. Since Windows prohibits path components from containing `:`
        # characters, failing to strip this `:`-suffixed prefix would
        # erroneously invalidate all valid absolute Windows pathnames.
        _, pathname = os.path.splitdrive(pathname)

        # Directory guaranteed to exist. If the current OS is Windows, this is
        # the drive to which Windows was installed (e.g., the "%HOMEDRIVE%"
        # environment variable); else, the typical root directory.
        root_dirname = (
            os.environ.get("HOMEDRIVE", "C:")
            if sys.platform == "win32"
            else os.path.sep
        )
        assert os.path.isdir(root_dirname)  # ...Murphy and her ironclad Law

        # Append a path separator to this directory if needed.
        root_dirname = root_dirname.rstrip(os.path.sep) + os.path.sep

        # Test whether each path component split from this pathname is valid or
        # not, ignoring non-existent and non-readable path components.
        for pathname_part in pathname.split(os.path.sep):
            try:
                os.lstat(root_dirname + pathname_part)
            # If an OS-specific exception is raised, its error code
            # indicates whether this pathname is valid or not. Unless this
            # is the case, this exception implies an ignorable kernel or
            # filesystem complaint (e.g., path not found or inaccessible).
            #
            # Only the following exceptions indicate invalid pathnames:
            #
            # * Instances of the Windows-specific "WindowsError" class
            #   defining the "winerror" attribute whose value is
            #   "ERROR_INVALID_NAME". Under Windows, "winerror" is more
            #   fine-grained and hence useful than the generic "errno"
            #   attribute. When a too-long pathname is passed, for example,
            #   "errno" is "ENOENT" (i.e., no such file or directory) rather
            #   than "ENAMETOOLONG" (i.e., file name too long).
            # * Instances of the cross-platform "OSError" class defining the
            #   generic "errno" attribute whose value is either:
            #   * Under most POSIX-compatible OSes, "ENAMETOOLONG".
            #   * Under some edge-case OSes (e.g., SunOS, *BSD), "ERANGE".
            except OSError as exc:
                if hasattr(exc, "winerror"):
                    if exc.winerror == ERROR_INVALID_NAME:
                        return False
                elif exc.errno in {errno.ENAMETOOLONG, errno.ERANGE}:
                    return False
    # If a "TypeError" exception was raised, it almost certainly has the
    # error message "embedded NUL character" indicating an invalid pathname.
    except TypeError:
        return False
    # If no exception was raised, all path components and hence this
    # pathname itself are valid. (Praise be to the curmudgeonly python.)
    else:
        return True
    # If any other exception was raised, this is an unrelated fatal issue
    # (e.g., a bug). Permit this exception to unwind the call stack.
    #
    # Did we mention this should be shipped with Python already?


def is_path_creatable(pathname: str) -> bool:
    """
    `True` if the current user has sufficient permissions to create the passed
    pathname; `False` otherwise.
    """
    # Parent directory of the passed path. If empty, we substitute the current
    # working directory (CWD) instead.
    dirname = os.path.dirname(pathname) or os.getcwd()
    return os.access(dirname, os.W_OK)


def is_path_exists_or_creatable(pathname: str) -> bool:
    """
    `True` if the passed pathname is a valid pathname for the current OS _and_
    either currently exists or is hypothetically creatable; `False` otherwise.

    This function is guaranteed to _never_ raise exceptions.
    """
    try:
        # To prevent "os" module calls from raising undesirable exceptions on
        # invalid pathnames, is_pathname_valid() is explicitly called first.
        return is_pathname_valid(pathname) and (
            os.path.exists(pathname) or is_path_creatable(pathname)
        )
    # Report failure on non-fatal filesystem complaints (e.g., connection
    # timeouts, permissions issues) implying this path to be inaccessible. All
    # other exceptions are unrelated fatal issues and should not be caught here.
    except OSError:
        return False


def is_path_sibling_creatable(pathname: str) -> bool:
    """
    `True` if the current user has sufficient permissions to create **siblings**
    (i.e., arbitrary files in the parent directory) of the passed pathname;
    `False` otherwise.
    """
    # Parent directory of the passed path. If empty, we substitute the current
    # working directory (CWD) instead.
    dirname = os.path.dirname(pathname) or os.getcwd()

    try:
        # For safety, explicitly close and hence delete this temporary file
        # immediately after creating it in the passed path's parent directory.
        with tempfile.TemporaryFile(dir=dirname):
            pass
        return True
    # While the exact type of exception raised by the above function depends on
    # the current version of the Python interpreter, all such types subclass the
    # following exception superclass.
    except EnvironmentError:
        return False


def is_path_exists_or_creatable_portable(pathname: str) -> bool:
    """
    `True` if the passed pathname is a valid pathname on the current OS _and_
    either currently exists or is hypothetically creatable in a cross-platform
    manner optimized for POSIX-unfriendly filesystems; `False` otherwise.

    This function is guaranteed to _never_ raise exceptions.
    """
    pathname = str(pathname)
    try:
        # To prevent "os" module calls from raising undesirable exceptions on
        # invalid pathnames, is_pathname_valid() is explicitly called first.
        return is_pathname_valid(pathname) and (
            os.path.exists(pathname) or is_path_sibling_creatable(pathname)
        )
    # Report failure on non-fatal filesystem complaints (e.g., connection
    # timeouts, permissions issues) implying this path to be inaccessible. All
    # other exceptions are unrelated fatal issues and should not be caught here.
    except OSError:
        return False



from __future__ import annotations

from dataclasses import dataclass

import pandas as pd

from bioTea.utils.errors import UnsupportedChip


@dataclass
class GeoPlatform:
    accession: str
    manufacturer: str
    title: str

    def __str__(self) -> str:
        return (
            f"GEO - Platform ({self.accession}) - {self.title} from {self.manufacturer}"
        )


@dataclass
class GeoSample:
    accession: str
    source: str
    extracted_molecule: str
    organism: str
    conditions: dict
    suppl_data_ftp: str
    suppl_data_local_path: str = None

    def __str__(self) -> str:
        return f"GEO - Sample ({self.accession}): {self.__dict__}"

    @classmethod
    def sample_from_xmldict(cls: GeoSample, raw_dict: dict) -> GeoSample:
        """Generate a new instance of the class from a dict parsed from an xml"""
        # Fail if the sample is more than one channel
        if (ch_count := int(raw_dict["Channel-Count"])) != 1:
            raise UnsupportedChip(f"Too many channels to parse ({ch_count})")

        if type((char := raw_dict["Channel"]["Characteristics"])) is str:
            conditions = {"type": char}
        else:
            # This is a list of small dicts. We fuse them together.
            def dictify(dict):
                return {dict["@tag"]: dict["#text"]}

            char = [dictify(x) for x in char]
            conditions = {}
            [conditions.update(x) for x in char]

        return cls(
            accession=raw_dict["@iid"],
            organism=raw_dict["Channel"]["Organism"]["#text"],
            source=raw_dict["Channel"]["Source"],
            extracted_molecule=raw_dict["Channel"]["Molecule"],
            conditions=conditions,
            suppl_data_ftp=raw_dict["Supplementary-Data"]["#text"],
        )


@dataclass
class GeoSeries:
    accession: str
    platform: GeoPlatform
    authors: list[str]
    samples: list[GeoSample]

    def __str__(self) -> str:
        authors = ", ".join(self.authors)
        samples = ", ".join([str(x) for x in self.samples])
        return f"GEO - Series ({self.accession}) from {authors}. Platform {self.platform} with {len(self.samples)} samples: {samples}"

    @property
    def design(self) -> str:
        design = {"Sample_id": [sample.accession for sample in self.samples]}
        for condition in self.samples[0].conditions:
            design[condition] = [
                sample.conditions[condition] for sample in self.samples
            ]

        return design

    @property
    def design_strings(self) -> str:
        strings = {}
        for key in self.design:
            strings[key] = ", ".join(self.design[key])
        return strings

    def generate_metadata(self) -> pd.DataFrame:
        return pd.DataFrame(data=self.design)



from __future__ import annotations

import json
import logging
import os
from abc import ABC
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from numbers import Number
from pathlib import Path
from typing import Any, Callable

import docker
import docker.errors
import requests
import typer
from docker.types import Mount
from packaging.version import LegacyVersion, parse
from typer import Abort

from bioTea import __version__
from bioTea.utils.errors import ContainerExitError, ImageNotFoundError
from bioTea.utils.path_checker import is_path_exists_or_creatable_portable
from bioTea.utils.tools import ConsoleWindow

log = logging.getLogger(__name__)

NAMESPACE = "cmalabscience"
LEAF = "biotea-box"
REPO = f"{NAMESPACE}/{LEAF}"

POSSIBLE_LOG_LEVELS = ("info", "debug", "error", "warning", "disable")


@dataclass
class BioTeaBoxVersion:
    raw_version: str

    @property
    def realversion(self):
        return parse(self.raw_version)

    def __lt__(self, other: BioTeaBoxVersion) -> bool:
        if type(other) is str:
            other = BioTeaBoxVersion(other)

        if type(other) is not BioTeaBoxVersion:
            raise TypeError(f"Cannot compare 'BioTeaBoxVersion' and '{type(other)}'")
        return self.realversion < other.realversion

    def __eq__(self, other: object) -> bool:
        if type(other) is str:
            other = BioTeaBoxVersion(other)

        if type(other) is not BioTeaBoxVersion:
            raise TypeError(f"Cannot compare 'BioTeaBoxVersion' and '{type(other)}'")
        return self.realversion == other.realversion

    def __str__(self) -> str:
        return self.raw_version


def get_installed_versions(
    client: docker.DockerClient = docker.from_env(),
) -> list[BioTeaBoxVersion]:
    local_images = client.images.list(REPO)
    local_images = [
        local_image.tags[0][(len(REPO) + 1) :] for local_image in local_images
    ]
    return [BioTeaBoxVersion(version) for version in local_images]


def pull_biotea_box_version(
    version: BioTeaBoxVersion, client: docker.DockerClient = docker.from_env()
):
    log.debug(f"Looking to see if {version} is available...")
    if not version in get_all_versions():
        log.error(f"Version {version} not found remotely.")
        raise ImageNotFoundError(str(version))

    log.info(f"Pulling remote image {version}...")
    client.images.pull(f"{REPO}:{version.raw_version}")
    log.info("Pulled image.")


def delete_biotea_box_version(
    version: BioTeaBoxVersion, client: docker.DockerClient = docker.from_env()
):
    if not version in get_installed_versions():
        log.error(f"Version {version} not found locally.")
        raise ImageNotFoundError(str(version))

    log.info(f"Removing image {version}...")
    client.images.remove(image=f"{REPO}:{version}")
    log.debug(f"Done removing image.")

    return True


def get_all_versions() -> list[BioTeaBoxVersion]:
    endpoint = f"https://registry.hub.docker.com/v2/namespaces/{NAMESPACE}/repositories/{LEAF}/tags"
    try:
        res = requests.get(endpoint, timeout=20)
    except Exception as e:
        log.exception("Failed to retrieve the biotea-box repo information", e)
        raise Abort

    images = json.loads(res.text)["results"]

    versions = []
    for image in images:
        version = BioTeaBoxVersion(image["name"])
        if type(version.realversion) is LegacyVersion:
            log.debug(f"Version '{version}' discarded as it is a Legacy version.")
            continue
        versions.append(version)

    if not versions:
        log.warn("No valid remote versions have been found.")

    return versions


def get_latest_version() -> BioTeaBoxVersion:
    all_vers = get_all_versions()
    if all_vers:
        return sorted(all_vers)[-1]
    raise ImageNotFoundError("No valid 'latest' version found.")


def is_version_compatible(version: BioTeaBoxVersion) -> bool:
    """Test if the input version is compatible with the current bioTEA version.

    Assumes compatibility for the future (bleeding) as well as all versions with
    identical major and minor versions.

    Args:
        version (BioTeaBoxVersion): The version to check for compatibility.

    Returns:
        bool: Whether the version is compatible or not.
    """

    if version == "bleeding":
        return True

    biotea_parsed = parse(__version__)

    log.debug(
        f"Checking for version compatibility."
        f"BioTEA version: {biotea_parsed}. Inputted version: {version.realversion}"
    )

    return (
        version.realversion.major == biotea_parsed.major
        and biotea_parsed.minor == version.realversion.minor
    )


# Some checks
def na_or(check) -> Callable:
    def _wrapped_check(argument):
        if argument is None:
            return True
        if type(argument) == str and argument.upper() in ("NA"):
            return True
        else:
            return check(argument)

    return _wrapped_check


def is_(argtype) -> Callable:
    def _wrapped_check(argument):
        return issubclass(type(argument), argtype)

    return _wrapped_check


def is_in(arglist) -> Callable:
    def _wrapped_check(argument):
        return argument in arglist

    return _wrapped_check


def is_valid_design_string(argument):
    # TODO: implement this check
    return is_(str)(argument)


def is_valid_color(argument):
    # TODO: implement this
    return is_(str)(argument)


def is_list_of(check) -> Callable:
    def _wrapped_check(argument):
        if not type(argument) == list:
            return False

        return all([check(x) for x in argument])

    return _wrapped_check


def _or_(x: Callable, y: Callable):
    def _wrapped(argument):
        return x(argument) or y(argument)

    return _wrapped


class BioTeaBoxArgument:
    """Class used to mark an argument in a BioTeaBoxInterface dict."""

    def __init__(self, check: Callable, default: Any) -> None:
        self.check = check
        self.default = default

        assert self.check(
            default
        ), "Invalid default BioTeaBox value. Someone coded it wrong."

    def __call__(self, argument=None):
        if argument == None:
            argument == self.default

        if not self.check(argument):
            raise ValueError(f"Argument check failed. Invalid argument {argument}")


class RequiredBioTeaBoxArgument(BioTeaBoxArgument):
    def __init__(self, check: Callable) -> None:
        self.check = check
        # The default does not matter. It HAS to be overridden.
        self.default = None

    def __call__(self, argument):
        if not self.check(argument):
            raise ValueError(f"Argument check failed. Invalid argument {argument}")


class BioTeaBoxInterface(ABC):
    """Abstract class that models an interface with BioTeaBox.

    Defines the arguments that can be passed to a BioTeaBox command, and handles
    parsing them to an object that can be given to `run_biotea_box` to run the
    concrete command.
    """

    possible_args: dict = None
    """The possible arguments to the interface.

    This is a dictionary of "value_name" = BioTeaBoxArgument.
    If a Callable, it is used to test the arg before passing it to BioTeaBox.
    If a RequiredArgument,
    """

    # I am not 100% sure this is the correct way to use this, but it fails
    # if "possible_args" is not defined, so I'm happy.
    def __init_subclass__(cls, /, **kwargs):
        super().__init_subclass__(**kwargs)
        if cls.possible_args is None:
            raise TypeError("Must specify `possible_args`.")

    @classmethod
    def parse_arguments(self, **kwargs) -> str:
        """Check the input args and parse them to a BioTeaBox-compliant string."""
        required_args = [
            key
            for key, val in self.possible_args.items()
            if type(val) is RequiredBioTeaBoxArgument
        ]

        assert all(
            [x in kwargs.keys() for x in required_args]
        ), "Missing required args: {}".format(
            ", ".join([x for x in required_args if x not in kwargs.keys()])
        )
        assert all(
            [x in self.possible_args.keys() for x in kwargs.keys()]
        ), "Unrecognized argument(s) {}".format(
            ", ".join([x for x in kwargs.keys() if x not in self.possible_args.keys()])
        )

        for key, value in kwargs.items():
            # These will raise an error if the check fails. So, if this passes,
            # all passed args are OK.
            try:
                self.possible_args[key](value)
            except ValueError:
                # Re-raise with more context
                raise ValueError(f"Argument check failed for key {key}: {value}")

        # Here, we are sure of three things:
        # 1. All required arguments are overridden by `kwargs`
        # 2. All arguments in `kwargs` are in the possible_arguments.
        # 3. All arguments in `kwargs` are valid (they pass the checks)
        # Therefore, we can update the default values with the kwargs safely.
        defaults = {
            key: self.possible_args[key].default for key in self.possible_args.keys()
        }
        defaults.update(kwargs)

        # BioTeaBox wants a JSON-encoded string.
        return f"'{json.dumps(defaults, indent=None)}'"


## >> Interface definitions
# To define a new interface, or check that one is compilant, look in the
# `entrypoint.R` of the BioTeaBox module, at the `defaults` list.
# Convert this list to a dictionary, replacing every `NULL` with RequiredArgument


class PrepAffyInterface(BioTeaBoxInterface):
    possible_args: dict = {
        "output.file": RequiredBioTeaBoxArgument(is_path_exists_or_creatable_portable),
        "remove.controls": BioTeaBoxArgument(is_(bool), True),
        "n_plots": BioTeaBoxArgument(is_(int), 1_000_000),
        # Plot options
        "use_pdf": BioTeaBoxArgument(is_(bool), True),
        "plot_width": BioTeaBoxArgument(is_(int), 16),
        "plot_height": BioTeaBoxArgument(is_(int), 9),
        "png_ppi": BioTeaBoxArgument(is_(int), 250),
        "enumerate_plots": BioTeaBoxArgument(is_(bool), True),
    }


class PrepAgilInterface(BioTeaBoxInterface):
    possible_args: dict = {
        "output_file": RequiredBioTeaBoxArgument(is_path_exists_or_creatable_portable),
        "remove_controls": BioTeaBoxArgument(is_(bool), True),
        "n_plots": BioTeaBoxArgument(is_(int), 1_000_000),
        "grep_pattern": BioTeaBoxArgument(is_(str), "*.(txt|TXT)"),
        # Plot options
        "use_pdf": BioTeaBoxArgument(is_(bool), True),
        "plot_width": BioTeaBoxArgument(is_(int), 16),
        "plot_height": BioTeaBoxArgument(is_(int), 9),
        "png_ppi": BioTeaBoxArgument(is_(int), 250),
        "enumerate_plots": BioTeaBoxArgument(is_(bool), True),
    }


class AnalyzeInterface(BioTeaBoxInterface):
    possible_args: dict = {
        "input.file": RequiredBioTeaBoxArgument(is_path_exists_or_creatable_portable),
        "experimental_design": RequiredBioTeaBoxArgument(is_valid_design_string),
        "contrasts": RequiredBioTeaBoxArgument(is_list_of(is_(str))),
        "min_log2_expression": BioTeaBoxArgument(is_(Number), 4.0),
        "fc_threshold": BioTeaBoxArgument(is_(Number), 0.5),
        "min_groupwise_presence": BioTeaBoxArgument(is_(Number), 0.8),
        "show_data_snippets": BioTeaBoxArgument(is_(bool), True),
        "annotation_database": BioTeaBoxArgument(is_(bool), True),
        "dryrun": BioTeaBoxArgument(is_(bool), False),
        "renormalize": BioTeaBoxArgument(is_(bool), False),
        "convert_counts": BioTeaBoxArgument(is_(bool), False),
        "run_limma_analysis": BioTeaBoxArgument(is_(bool), True),
        "run_rankprod_analysis": BioTeaBoxArgument(is_(bool), True),
        "batches": BioTeaBoxArgument(na_or(is_list_of(is_valid_design_string)), "NA"),
        "extra_limma_vars": BioTeaBoxArgument(
            na_or(is_list_of(is_valid_design_string)), "NA"
        ),
        "group_colors": BioTeaBoxArgument(
            is_list_of(is_valid_color),
            [
                "cornflowerblue",
                "firebrick3",
                "olivedrab3",
                "darkgoldenrod1",
                "purple",
                "magenta3",
            ],
        ),
        # Plot options
        "use_pdf": BioTeaBoxArgument(is_(bool), True),
        "plot_width": BioTeaBoxArgument(is_(int), 16),
        "plot_height": BioTeaBoxArgument(is_(int), 9),
        "png_ppi": BioTeaBoxArgument(is_(int), 250),
        "enumerate_plots": BioTeaBoxArgument(is_(bool), True),
    }


class AnnotateInterface(BioTeaBoxInterface):
    possible_args: dict = {
        "expression_data_path": RequiredBioTeaBoxArgument(
            is_path_exists_or_creatable_portable
        ),
        "output_path": RequiredBioTeaBoxArgument(is_path_exists_or_creatable_portable),
        "database_name": BioTeaBoxArgument(_or_(is_(str), is_(bool)), "internal"),
    }


## <<


def run_biotea_box(
    command: str,
    arguments: dict[str],
    interface: BioTeaBoxInterface,
    input_anchor: Path,
    output_anchor: Path,
    log_anchor: Path,
    version: str = "latest",
    console_level: str = "info",
    logfile_level: str = "debug",
    log_name: str = "auto",
) -> int:
    assert logfile_level in POSSIBLE_LOG_LEVELS
    assert console_level in POSSIBLE_LOG_LEVELS

    client = docker.from_env()

    if version == "latest":
        version = get_latest_version()
        log.info(f"Latest version: {version}")

    if version not in get_installed_versions(client=client):
        pull_biotea_box_version(version, client=client)

    if not is_version_compatible(version):
        log.warn(
            f"Selected an incompatible version '{version}'. BioTEA might not work."
        )
        typer.confirm("Continue anyway?", abort=True)

    for path in [input_anchor, output_anchor, log_anchor]:
        assert is_path_exists_or_creatable_portable(
            path
        ), f"Path {path} is inaccessible."
        assert not path.is_file(), f"Path {path} points to a file, not a folder."

        if not path.exists():
            log.debug(f"Making anchor point: {path}")
            os.makedirs(path)

    try:
        image = client.images.get(f"{REPO}:{version}")
    except docker.errors.ImageNotFound:
        log.error(f"Cannot find local image: {version}")
        return 0

    if log_name == "auto":
        now = datetime.today().strftime("%Y-%m-%d-%H:%M:%S")
        log_name = f"bioTEA_{now}"

    log.debug(f"Parsing arguments with interface '{type(interface)}'")
    try:
        parsed_args = interface.parse_arguments(**arguments)
    except ValueError:
        log.exception(
            "Invalid arguments passed to interface. Please open an issue with the bioTEA logs."
        )
        return 0

    log.info(f"Launching container with version {version}")

    try:
        # The composed command is like such:
        # UUID GUID command logname loglevel_console loglevel_file (args)
        composed_command = (
            str(os.getuid())
            + " "
            + str(os.getgid())
            + " "
            + command
            + " "
            + log_name
            + " "
            + console_level
            + " "
            + logfile_level
            + " "
            + parsed_args
        )
        log.info(f"Composed command: {composed_command}")
        container = client.containers.run(
            image,
            command=composed_command,
            stderr=True,
            detach=True,
            mounts=[
                # Needs the `str` or the internal serializer dies
                Mount("/bioTEA/target", str(output_anchor.absolute()), type="bind"),
                Mount(
                    "/bioTEA/input",
                    str(input_anchor.absolute()),
                    type="bind",
                    read_only=True,
                ),
                Mount("/bioTEA/logs", str(log_anchor.absolute()), type="bind"),
            ],
        )
    except Exception as e:
        log.error(f"Launching container failed: {e}")
        raise e

    try:
        with ConsoleWindow(10, name="BioTeaBox container", line_prefix="> ") as window:
            for line in container.logs(stream=True):
                window.print(line.decode().rstrip())
    except KeyboardInterrupt:
        log.warn("Got shutdown signal. Killing container.")
        container.kill()
        container.remove()
        raise KeyboardInterrupt

    try:
        # Sometimes the container does not stop immediately.
        # This assures it does, even on errors.
        container.stop()
    except Exception:
        pass

    # Reload the container status
    container.reload()
    statuscode = container.wait()["StatusCode"]
    log.debug(f"Container exited with status {container.status} [{statuscode}]")
    log.debug(f"Removing container...")
    container_logs = container.logs().decode()
    container.remove()
    if statuscode != 0:
        container_error = "\n".join(container_logs.split("\n")[-5:])
        raise ContainerExitError(
            f"Container had a non-zero exit status: {statuscode}. Last five lines of container output: {container_error}"
        )

    return 0


class SpecialCommand(Enum):
    versions = "versions"
    test = "test"


def run_special_biotea_command(command: SpecialCommand, version: BioTeaBoxVersion):
    """Runs a special bioTEA box command that does not require args.

    This means we also do not need an interface, or mounts, or anything special.
    """
    client = docker.from_env()

    if version == "latest":
        version = get_latest_version()
        log.info(f"Latest version: {version}")

    if version not in get_installed_versions(client=client):
        pull_biotea_box_version(version, client=client)

    if not is_version_compatible(version):
        log.warn(
            f"Selected an incompatible version '{version}'. BioTEA might not work."
        )
        typer.confirm("Continue anyway?", abort=True)

    try:
        image = client.images.get(f"{REPO}:{version}")
    except docker.errors.ImageNotFound:
        log.error(f"Cannot find local image: {version}")
        return 0

    log.info(
        f"Running special command '{command.value}' in container version '{version}'."
    )

    composed_command = str(os.getuid()) + " " + str(os.getgid()) + " " + command.value

    bin_logs = client.containers.run(image, command=composed_command, stderr=True)

    print(bin_logs.decode("UTF-8"))

    return 0



from __future__ import annotations

import collections
import ftplib
import gzip
import logging
import os
import sys
import tarfile
import zipfile
from collections import deque
from copy import copy
from math import floor
from pathlib import Path
from shutil import get_terminal_size
from typing import Callable, Optional, Union

import requests
import typer
import yaml
from tqdm import tqdm
from tqdm.utils import CallbackIOWrapper

from bioTea.utils.errors import InvalidGeoId, InvalidPathError
from bioTea.utils.path_checker import is_pathname_valid

log = logging.getLogger(__name__)

# PathLike: TypeAlias = Union[Path, str]
class PathLike:
    pass


CompressedFile = Union[tarfile.TarFile, zipfile.ZipFile, gzip.GzipFile]


class TempWorkdir:
    """Change the workdir temporarily.

    Probably super unsafe for threading.
    """

    def __init__(self, tempwd: PathLike) -> None:
        tempwd = Path(tempwd).expanduser().resolve()
        if not tempwd.exists():
            os.makedirs(tempwd)

        self.last_wd = os.getcwd()
        self.new_wd = tempwd

    def __enter__(self):
        os.chdir(self.new_wd)

    def __exit__(self, exc_type, exc_val, exc_tb):
        os.chdir(self.last_wd)


def ask_user(
    prompt: str = "Continue?",
    options: dict = {"yes": lambda: None, "no": lambda: sys.exit()},
    allow_partial: bool = True,
    ignore_case: bool = True,
) -> None:
    """Ask the user a question with a number of options.

    Execute a fuction based on the answer given. Returns the value of the function ran.

    Args:
        prompt (str): The prompt that the user sees.
        options (dict, optional): A dict with keys as the possible answers and as values callables that will be called when the function is ran. Defaults to {"yes": lambda: None, "no": lambda: sys.exit()}.
        allow_partial (bool, optional): Allow partial answers (detected by startswith - careful with keys that start the same)?. Defaults to True.
        ignore_case (bool, optional): Ignore case of options and input?. Defaults to True.
    """
    if ignore_case:
        options = dict((k.lower(), v) for k, v in options.iteritems())

    prompt = "{} [{}]:".format(prompt, "/".join(options.keys()))

    while True:
        response = input(prompt)
        if ignore_case:
            response = response.lower()

        if allow_partial:
            for k, v in options.iteritems():
                if k.startswith(response):
                    return v()
        else:
            for k, v in options.iteritems():
                if k == response:
                    return v()

        print("Please reply with one of '{}'".format(" ,".join(options.keys())))


def user_input(
    prompt: str,
    test: Callable,
    retry: bool = True,
    retry_prompt: str = f"Invalid input. Please try again.",
) -> str:
    """Ask for input from the user.

    Tests if the answer is valid with a function, and can be made to retry until the function passes.

    Args:
        prompt (str): The prompt shown to the user at every iteration.
        test (Callable): The test applied to the input. Needs to return a bool.
        retry (bool, optional): Should the function retry? If false, raises a ValueError if the test did not pass. Defaults to True.
        retry_prompt (str, optional): Text to show before giving the prompt again to the user. Defaults to f"Invalid input. Please try again.".

    Raises:
        ValueError: If the test fails and `retry` is not true.

    Returns:
        str: The (tested) user input.
    """
    while True:
        raw_input = input(prompt)

        if test(raw_input.strip()):
            return raw_input

        if not retry:
            raise ValueError(
                f"Invalid check {test} for input {raw_input}, and cannot retry."
            )

        print(retry_prompt)


def download_ftp(
    ftp_url: str, destination: PathLike, filter: Callable = lambda x: True
) -> dict:
    """Downloads a file or a set of files from a ftp url

    Args:
        ftp_url: A string. May or may not start with `ftp://`
        destination: PathLike. The destination folder.
        filter: A function that takes as string. Applied to all filenames to
            download. Will dowload only those that make the function return true.

    Returns:
        A dict with keys the names of the downloaded paths and values the
        paths to the files on disk
    """
    destination = Path(destination)
    result = {}

    if not destination.exists():
        log.warn(f"Making destination: {destination}")
        os.makedirs(destination)

    if ftp_url.startswith("ftp://"):
        ftp_url = ftp_url[6:]
    host = ftp_url.split("/")[0]
    path = "/".join(ftp_url.split("/")[1:])

    log.info("Connecting to remote server....")
    log.debug(f"Connecting to '{host}'...")
    ftp = ftplib.FTP(host, timeout=240)
    log.debug(f"Logging in to '{host}'...")
    ftp.login()

    log.debug(f"Moving to '{path}'")
    ftp.cwd(path)

    files = []
    ftp.retrlines("LIST", callback=files.append)
    files = [x.split()[-1] for x in files]

    if not files:
        log.info("No files to download.")
        return {}

    all_files = len(files)
    files = [file for file in files if filter(file)]

    log.info(
        "Found {} files: {} ({} filtered out)".format(
            len(files), ", ".join(files), all_files - len(files)
        )
    )

    if not files:
        log.warning("Filter function removed all files. Nothing to do.")
        return {}

    destinations = [destination / file for file in files]

    log.debug(f"Starting download of {destinations}")
    for file, destination in tqdm(zip(files, destinations)):
        with destination.open("wb+") as outfile, tqdm(
            unit="B", unit_scale=True, unit_divisor=1024
        ) as pbar:
            outfile = CallbackIOWrapper(pbar.update, outfile, "write")

            def cb(data):
                outfile.write(data)

            ftp.retrbinary(f"RETR {file}", cb)

        result.update({file: destination})

    log.info("Finished downloading files.")
    return result


def ping(url: str) -> bool:
    """Ping some url and return true if it exist"""
    if not url.startswith("http://") or url.startswith("https://"):
        url = f"http://{url}"
    response = requests.get(url)
    return response.status_code == 200


# {obsid} is the obscured id, with the last three chars substituted by "nnn":
# e.g.: GPL1234 |> GPL1nnn
# {id} is the actual id
GEO_ENDPOINTS = {
    "GPL": {
        "miniml": "ftp://ftp.ncbi.nlm.nih.gov/geo/platforms/{obsid}/{id}/miniml/",
        "soft": "ftp://ftp.ncbi.nlm.nih.gov/geo/platforms/{obsid}/{id}/soft/",
        "suppl": "ftp://ftp.ncbi.nlm.nih.gov/geo/platforms/{obsid}/{id}/suppl/",
    },
    "GDS": {
        "miniml": "ftp://ftp.ncbi.nlm.nih.gov/geo/datasets/{obsid}/{id}/miniml/",
        "soft": "ftp://ftp.ncbi.nlm.nih.gov/geo/datasets/{obsid}/{id}/soft",
        "suppl": "ftp://ftp.ncbi.nlm.nih.gov/geo/datasets/{obsid}/{id}/suppl",
    },
    "GSM": {
        "miniml": "ftp://ftp.ncbi.nlm.nih.gov/geo/samples/{obsid}/{id}/miniml/",
        "soft": "ftp://ftp.ncbi.nlm.nih.gov/geo/samples/{obsid}/{id}/soft",
        "suppl": "ftp://ftp.ncbi.nlm.nih.gov/geo/samples/{obsid}/{id}/suppl",
    },
    "GSE": {
        "miniml": "ftp://ftp.ncbi.nlm.nih.gov/geo/series/{obsid}/{id}/miniml/",
        "soft": "ftp://ftp.ncbi.nlm.nih.gov/geo/series/{obsid}/{id}/soft",
        "suppl": "ftp://ftp.ncbi.nlm.nih.gov/geo/series/{obsid}/{id}/suppl",
    },
}


def make_geo_ftp(geo_id: str, type: str = "miniml") -> str:
    """Get a geo FTP url from an id.

    Type can be 'miniml', 'soft' or 'suppl'

    Args:
        geo_id (str): The geo id to format with.
        type (str, optional): The type of file needed. Can be "miniml", "soft" or "suppl" for minimal files, soft files or supplementary files, respectively. Defaults to "miniml".

    Raises:
        InvalidGeoId: If the GEO id cannot be used to generate an ftp url.

    Returns:
        str: A string with the GEO ftp url pointing to the right destination.
    """
    assert type in ["miniml", "soft", "suppl"]
    geo_id = geo_id.upper()
    obsid = f"{geo_id[:-3]}nnn"

    try:
        endpoints = GEO_ENDPOINTS[geo_id[:3]]
    except KeyError:
        raise InvalidGeoId(f"Invalid ID root '{geo_id[:3]}'.")

    return endpoints[type].format(obsid=obsid, id=geo_id)


def contains_all(x: list, y: list) -> bool:
    if len(x) != len(y):
        return False

    return all(val in y for val in x)


class ConsoleWindow:
    def __init__(
        self,
        height: int,
        name: Optional[str] = None,
        line_prefix: str = "",
        clear: bool = False,
    ) -> None:
        self.height = height
        self.name = name
        self.line_prefix = line_prefix
        self.clear = clear

        self.buffer = deque([""] * self.height, maxlen=self.height)

    def print(self, line) -> None:
        max_chars, _ = get_terminal_size((80, 20))
        max_chars = floor(max_chars * 0.9)  # Give it some wiggle space
        if len(line) > max_chars:
            lines = [
                line[i : i + max_chars].strip() for i in range(0, len(line), max_chars)
            ]
        else:
            lines = [line.strip()]

        self.buffer.extend(lines)
        self.update()

    @staticmethod
    def clearlines(n):
        # Use ANSI codes to clear `n` lines
        for _ in range(n):
            # The first goes up one line, the second clears it.
            print("\033[1A", end="\x1b[2K")

    def update(self) -> None:
        self.clearlines(self.height)
        for line in self.buffer:
            print(self.line_prefix + line)

    def __enter__(self) -> ConsoleWindow:
        if self.name:
            print(self.line_prefix + f"-- {self.name} -- " + "\n" * self.height)
        else:
            print(self.line_prefix + "\n" * self.height)

        return self

    def __exit__(self, type, value, traceback) -> None:
        self.update()

        if self.clear:
            if self.name:
                self.clearlines(self.height + 1)
            else:
                self.clearlines(self.height)


def make_path_valid(path: Path, dir: bool = False):
    """Make a path valid to write in

    If the path points to a real directory or file, makes sure that we can read
    and write there. Otherwise, makes the directory/parent directory so we
    can use the path directly.

    Args:
        path (Path): The path to test
        dir (bool, optional): Is the path theoretically pointing to  a directory? Defaults to False.

    Raises:
        InvalidPathError: If the path is invalid.

    Returns:
        Path: The same path in input.
    """
    path = path.expanduser().resolve()
    if is_pathname_valid(str(path)) and not path.exists():
        if dir:
            os.makedirs(path, exist_ok=True)
        else:
            os.makedirs(path.parent, exist_ok=True)

    if not is_pathname_valid(str(path)):
        raise InvalidPathError(f"Invalid path: {path}")

    return path


def parse_biotea_box_options(path: Path) -> dict:
    with path.open("r") as stream:
        raw_args = yaml.safe_load(stream)

    # I need to parse the raw dictionary to a shallow input redable by the box
    # The options that can be accepted are in `docker_wrapper.py/AnalizeInterface`,
    # which in turn looks to the entrypoint of the analize module.
    try:
        args = {
            "experimental_design": raw_args["design"]["experimental_design"],
            "contrasts": raw_args["design"]["contrasts"],
            "min_log2_expression": raw_args["design"]["filters"]["log2_expression"],
            "fc_threshold": raw_args["design"]["filters"]["fold_change"],
            "min_groupwise_presence": raw_args["design"]["filters"][
                "min_groupwise_presence"
            ],
            "show_data_snippets": raw_args["general"]["show_data_snippets"],
            "annotation_database": raw_args["general"]["annotation_database"],
            "dryrun": raw_args["switches"]["dryrun"],
            "renormalize": raw_args["switches"]["renormalize"],
            "convert_counts": raw_args["switches"]["convert_counts"],
            "run_limma_analysis": raw_args["switches"]["limma"],
            "run_rankprod_analysis": raw_args["switches"]["rankproduct"],
            "group_colors": raw_args["design"]["group_colors"],
            # Plot options
            "use_pdf": not raw_args["general"]["plots"]["save_png"],
            "plot_width": raw_args["general"]["plots"]["plot_width"],
            "plot_height": raw_args["general"]["plots"]["plot_height"],
            "png_ppi": raw_args["general"]["plots"]["png_resolution"],
            "enumerate_plots": raw_args["general"]["plots"]["enumerate_plots"],
        }
    except KeyError as e:
        log.error(f"Cannot find a valid option for key {e.args[0]}. Aborting.")
        sys.exit()

    # These two options can be null, so a missing key is fine:
    try:
        batches = raw_args["design"]["batches"]
    except KeyError:
        batches = "NA"
    try:
        extra_limma_vars = raw_args["batches"]["extra_limma_vars"]
    except KeyError:
        extra_limma_vars = "NA"

    args.update({"batches": batches, "extra_limma_vars": extra_limma_vars})

    return args


def contains_numbers(some_string):
    # If i get a number, need to make it a string.
    some_string = str(some_string)
    return any(char.isdigit() for char in some_string)


def infinite(iterable):
    original = copy(iterable)
    times = 1
    while True:
        for item in iterable:
            yield item * times
        iterable = copy(original)
        times += 1


class Replacer:
    def __init__(self, replacements) -> None:
        self.replacements = replacements
        self.iterator = infinite(iter(replacements))
        self.matches = {}

    def sanitize(self, values: list) -> list:
        res = []
        for value in values:
            try:
                replacement = self.matches[value]
            except KeyError:
                replacement = next(self.iterator)
                self.matches[value] = replacement
            res.append(replacement)

        return res


def ask_choices(
    prompt: str, choices: list, accept_list: bool = False
) -> Union[str, list[str]]:
    if "" in choices:
        clear_choices = [x for x in choices if x != ""]
        composed_prompt = f"{prompt} " + ", ".join(clear_choices) + " or leave empty: "
    else:
        composed_prompt = f"{prompt} " + ", ".join(choices) + ": "

    while True:
        user = input(composed_prompt)
        if accept_list:
            user = [x.strip() for x in user.split(",")]
            if not all([x in choices for x in user]):
                invalid = [x for x in user if x not in choices]
                typer.echo(f"Invalid choice(s): {invalid}. Try again.")
                continue
            return user
        else:
            if user not in choices:
                typer.echo("Invalid choice: {user}. Try again.")
                continue
            return user


def recursive_dict_update(d, u):
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = recursive_dict_update(d.get(k, {}), v)
        else:
            d[k] = v
    return d
